# SPDX-FileCopyrightText: © 2024 Tenstorrent AI ULC
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

__all__ = ["CoreId", "TtxFile", "load_ttx_file", "TtxCompletionChecks"]

from dataclasses import dataclass
import itertools
import re
import struct
import zipfile
from typing import (
    Optional,
    Iterable,
    Tuple,
    AbstractSet,
    Mapping,
    NamedTuple,
    Dict,
    List,
    Any,
    Collection,
    AnyStr,
    IO,
    BinaryIO,
    ClassVar,
)

from yaml import safe_load

from tt_burnin.chip import BhChip, TTChip as Chip

DEFAULT_TIMEOUT_CYCLES = 100_000


# [0] = x, [1] = y
class CoreId(NamedTuple):
    x: int
    y: int

    def __str__(self) -> str:
        return f"{self.x}-{self.y}"

    @classmethod
    def parse(cls, text: AnyStr) -> "CoreId":
        m = re.fullmatch(r"(\d+)-(\d+)", str(text))
        if m is None:
            raise ValueError("Could not parse core id.")
        return cls(int(m[1]), int(m[2]))


class TtxFile(zipfile.ZipFile):
    def __init__(self, file, mode: str = "r"):
        super().__init__(file, mode)
        self._test_yaml = None

    def testdef(self) -> Any:
        if self._test_yaml is None:
            self._test_yaml = safe_load(self.open("test.yaml"))

        return self._test_yaml


BIN_FILE_V1_MAGIC = 0x9704266B
BIN_HEADER_STRUCT = struct.Struct("<IIIIIIII")


@dataclass
class ChunkHeader:
    address: int
    length: int
    reserved_mbz: int = 0

    CHUNK_HEADER_STRUCT: ClassVar = struct.Struct("<QII")

    def write(self, f: BinaryIO) -> None:
        f.write(
            ChunkHeader.CHUNK_HEADER_STRUCT.pack(
                self.address, self.length, self.reserved_mbz
            )
        )

    @staticmethod
    def read(f: IO[bytes]) -> Optional["ChunkHeader"]:
        b = f.read(ChunkHeader.CHUNK_HEADER_STRUCT.size)
        if len(b) == 0:
            return None
        if len(b) != ChunkHeader.CHUNK_HEADER_STRUCT.size:
            raise RuntimeError("Image file is truncated within chunk header.")

        ch = ChunkHeader.CHUNK_HEADER_STRUCT.unpack(b)
        if ch[2] != 0:
            raise RuntimeError("Chunk header contains nonzero in MBZ field.")

        return ChunkHeader(*ch)


def check_bin_header(f: IO[bytes]) -> None:
    b = f.read(BIN_HEADER_STRUCT.size)
    if len(b) != BIN_HEADER_STRUCT.size:
        raise RuntimeError("Image file is truncated within binary file header.")

    h = BIN_HEADER_STRUCT.unpack(b)

    if h[0] != BIN_FILE_V1_MAGIC:
        raise RuntimeError("Image file does not start with expected magic.")

    if any(map(lambda x: x != 0, h[1:])):
        raise RuntimeError("Image file header contains nonzero in MBZ field.")


def read_bin_image_chunks(f: IO[bytes]) -> Iterable[Tuple[int, bytes]]:
    check_bin_header(f)

    while True:
        chunk_header = ChunkHeader.read(f)
        if chunk_header is None:
            break

        chunk_data = f.read(chunk_header.length)
        if len(chunk_data) != chunk_header.length:
            raise RuntimeError("Image file is truncated within data chunk.")

        yield (chunk_header.address, chunk_data)


def read_hex_image_chunks(f: IO[bytes]) -> Iterable[Tuple[int, bytes]]:
    buffer = bytearray()
    addr = None
    for line in f.readlines():
        if len(line.strip()) > 0:
            if line.startswith(b"@"):
                if len(buffer) > 0:
                    assert addr is not None
                    yield addr, buffer
                # Word addr, so multiply it by 4 to turn it into byte address
                addr = int(f"0x{line[1:].strip().decode()}", 0) * 4
                buffer = bytearray()
            else:
                buffer.extend(
                    int(f"0x{line.strip().decode()}", 0).to_bytes(4, "little")
                )

    if len(buffer) > 0:
        assert addr is not None
        yield addr, buffer


def load_hex(chip: Chip, cores: Optional[Collection[CoreId]], bin: IO[bytes]) -> None:
    for address, data in read_hex_image_chunks(bin):
        if cores is None:
            chip.noc_broadcast(0, address, data)
        else:
            for core in cores:
                chip.noc_write(0, *core, address, data)


def check_hex(chip: Chip, cores: Optional[Collection[CoreId]], bin: IO[bytes]) -> None:
    for address, data in read_hex_image_chunks(bin):
        if cores is None:
            cores = chip.get_tensix_locations()
        for core in cores:
            buffer = bytearray(len(data))
            chip.noc_read(0, *core, address, buffer)
            if buffer != data:
                for b, d in zip(buffer, data):
                    assert (
                        b == d
                    ), f"Failed to write to core {address} {core} ({b} != {d})"


def load_bin(chip: Chip, cores: Optional[Collection[CoreId]], bin: IO[bytes]) -> None:
    for address, data in read_bin_image_chunks(bin):
        if cores is None:
            chip.noc_broadcast(0, address, data)
        else:
            for core in cores:
                chip.noc_write(0, *core, address, data)

def check_bin(chip: Chip, cores: Optional[Collection[CoreId]], bin: IO[bytes]) -> None:
    for address, data in read_bin_image_chunks(bin):
        if cores is None:
            cores = chip.get_tensix_locations()
        for core in cores:
            buffer = bytearray(len(data))
            chip.noc_read(0, *core, address, buffer)
            if buffer != data:
                for b, d in zip(buffer, data):
                    assert (
                        b == d
                    ), f"Failed to write to core {core} at address {address} ({b} != {d})"

# core_mapping: use logical (source) to physical (target) mapping
# Returns the physical cores that it loaded an image on to.
def load_ttx_file(
    chip: Chip,
    ttx: TtxFile,
    core_mapping: Mapping[CoreId, Collection[CoreId]],
    no_check: bool,
) -> AbstractSet[CoreId]:
    all_tensix_cores = set(CoreId(*c) for c in chip.get_tensix_locations())

    all_source_cores = set(core_mapping.keys())
    all_target_cores = set(itertools.chain.from_iterable(core_mapping.values()))

    if len(all_target_cores - all_tensix_cores) > 0:
        details = ", ".join(map(str, sorted(all_target_cores - all_tensix_cores)))
        raise RuntimeError(f"core_mapping targets cores that do not exist. ({details})")

    broadcast = (
        set(core_mapping.keys()) == set([CoreId(0, 0)])
        and set(core_mapping[CoreId(0, 0)]) == all_tensix_cores
    )

    # find all X-Y/(image|ckernels).(bin|hex)
    # If non image.bin, fail.
    # If any hex, fail.
    # If any core has ckernels.bin but not image.bin, fail.
    # If broadcast or single-core, and any cores other than 0-0, fail.
    # If neither broadcast nor single-core verify that all cores are in chip.get_tensix_locations()

    infolist = {info.filename: info for info in ttx.infolist()}

    files: Dict[str, set] = {
        "image.bin": set(),
        "image.hex": set(),
        "ckernels.bin": set(),
        "ckernels.hex": set(),
    }

    for info in infolist.values():
        m = re.fullmatch(r"(\d+)-(\d+)/((?:image|ckernels)\.(bin|hex))", info.filename)
        if m:
            x = int(m[1])
            y = int(m[2])

            image_type = m[4]

            # ignore empty images, they may exist for non-tensix cores
            if (image_type == "bin" and info.file_size > BIN_HEADER_STRUCT.size) or (
                image_type == "hex" and info.file_size > 0
            ):
                files[m[3]].add(CoreId(x, y))

    # Ignore hexs that are shadowed by bins. These existed for back-compat with old loaders.
    files["image.hex"] -= files["image.bin"]
    files["ckernels.hex"] -= files["ckernels.bin"]

    image_bins = files["image.bin"]
    image_hex = files["image.hex"]
    ckernels_bins = files["ckernels.bin"]
    ckernels_hex = files["ckernels.hex"]

    del files

    if not image_bins and not image_hex:
        raise RuntimeError("TTX is empty.")

    if len(ckernels_bins.union(ckernels_hex) - image_bins.union(image_hex)) > 0:
        details = ", ".join(map(str, sorted(ckernels_bins - image_bins)))
        raise RuntimeError(f"TTX has cores with ckernels but no image. ({details})")

    if len(image_bins.union(image_hex) - all_source_cores) > 0:
        details = ", ".join(
            map(str, sorted(image_bins.union(image_hex) - all_source_cores))
        )
        raise RuntimeError(
            f"TTX has images for cores with no physical mapping. ({details})"
        )

    def load_core(
        load_core: CoreId, target_cores: Optional[Collection[CoreId]], no_check: bool
    ) -> None:
        image_bin = f"{load_core}/image.bin"
        ckernels_bin = f"{load_core}/ckernels.bin"

        image_hex = f"{load_core}/image.hex"
        ckernels_hex = f"{load_core}/ckernels.hex"

        if image_hex in infolist:
            load_hex(chip, target_cores, ttx.open(infolist[image_hex], mode="r"))
            if not no_check:
                check_hex(chip, target_cores, ttx.open(infolist[image_hex], mode="r"))

        if ckernels_hex in infolist:
            load_hex(chip, target_cores, ttx.open(infolist[ckernels_hex], mode="r"))
            if not no_check:
                check_hex(
                    chip, target_cores, ttx.open(infolist[ckernels_hex], mode="r")
                )

        if image_bin in infolist:
            load_bin(chip, target_cores, ttx.open(infolist[image_bin], mode="r"))
            if not no_check:
                check_bin(chip, target_cores, ttx.open(infolist[image_bin], mode="r"))

        if ckernels_bin in infolist:
            load_bin(chip, target_cores, ttx.open(infolist[ckernels_bin], mode="r"))
            if not no_check:
                check_bin(
                    chip, target_cores, ttx.open(infolist[ckernels_bin], mode="r")
                )

    if broadcast:
        load_core(CoreId(0, 0), None, no_check)
        return all_tensix_cores

    else:
        for source_core in image_bins:
            load_core(source_core, core_mapping[source_core], no_check)
        return image_bins


@dataclass(frozen=True)
class AddressData:
    address: int
    data: bytes

    @classmethod
    def load_from_hex_file(cls, f: IO) -> "AddressData":
        line = f.readline()
        if len(line) == 0:
            raise RuntimeError("Empty memory file.")

        if line[0] != ord("@"):
            raise RuntimeError("Memory file does not start with address line.")

        address = int(line[1:], 16) * 4
        data = b"".join(map(lambda line: int(line, 16).to_bytes(4, "little"), f))

        return cls(address, data)


class TtxCompletionChecks:
    def __init__(self, ttx: Optional[TtxFile] = None):
        if ttx is not None:
            self._checks = self._load_from_ttx(ttx)
        else:
            self._checks = {}

    def _load_from_ttx(self, ttx: TtxFile) -> Dict[CoreId, List[AddressData]]:
        completion_checks: Dict[CoreId, List[AddressData]] = {}

        for item in ttx.testdef()["completion"]["filematches"]:
            core = CoreId(int(item["node"]["x"]), int(item["node"]["y"]))
            content = AddressData.load_from_hex_file(ttx.open(item["file"]))

            completion_checks.setdefault(core, []).append(content)

        return completion_checks

    # Create TtxCompletionChecks using pre-loaded (x,y,file) tuples.
    @classmethod
    def preloaded(
        cls, ttx: TtxFile, checks: Collection[Tuple[int, int, str]]
    ) -> "TtxCompletionChecks":
        completion_checks = cls()

        for x, y, file in checks:
            core = CoreId(x, y)
            content = AddressData.load_from_hex_file(ttx.open(file))
            completion_checks._checks.setdefault(core, []).append(content)

        return completion_checks

    # This merges checks for all cores, matching the original C++ code.
    def remap_for_broadcast(self, cores: Collection[CoreId]) -> "TtxCompletionChecks":
        all_checks = list(itertools.chain.from_iterable(self._checks.values()))

        new = type(self)()
        new._checks = {core: all_checks.copy() for core in cores}
        return new

    def remap_to_physical(
        self, core_mapping: Mapping[CoreId, Collection[CoreId]]
    ) -> "TtxCompletionChecks":
        new = type(self)()
        new._checks = {
            physical: self._checks[logical].copy()
            for logical, physicals in core_mapping.items()
            for physical in physicals
            if logical in self._checks
        }
        return new

    def filter_physical(
        self, loaded_cores: AbstractSet[CoreId]
    ) -> "TtxCompletionChecks":
        new = type(self)()
        new._checks = {
            core: checks.copy()
            for core, checks in self._checks.items()
            if core in loaded_cores
        }
        return new

    def empty(self) -> bool:
        return not bool(self._checks)

    def _read_block(self, chip: Chip, core: CoreId, address: int, length: int) -> bytes:
        alignment_offset = address % 16
        read_address = address - alignment_offset

        read_length = length + alignment_offset
        read_rounding = 16 - (read_length % 16)
        read_length += read_rounding if read_rounding < 16 else 0

        buffer = bytearray(read_length)
        chip.noc_read(0, *core, read_address, buffer)

        return buffer[alignment_offset : alignment_offset + length]

    def test(self, chip: Chip) -> bool:
        for core, checks in self._checks.items():
            for check in checks:
                data = self._read_block(chip, core, check.address, len(check.data))
                if data != check.data:
                    return False
        return True



# SPDX-FileCopyrightText: © 2024 Tenstorrent AI ULC
# SPDX-License-Identifier: Apache-2.0
from typing import List
import sys
import os
import json
import jsons
import time
import random
from rich.table import Table

from rich import get_console
from pyluwen import PciChip
from tt_tools_common.ui_common.themes import CMD_LINE_COLOR
from tt_tools_common.reset_common.bh_reset import BHChipReset
from tt_tools_common.reset_common.wh_reset import WHChipReset
from tt_tools_common.reset_common.gs_tensix_reset import GSTensixReset
from tt_tools_common.reset_common.galaxy_reset import GalaxyReset
from tt_tools_common.utils_common.tools_utils import (
    detect_chips_with_callback,
)
from pyluwen import (
    detect_chips_fallible,
    run_wh_ubb_ipmi_reset,
    run_ubb_wait_for_driver_load
)
from tt_burnin.chip import RemoteWhChip, WhChip


def pci_board_reset(list_of_boards: List[int], reinit=False):
    """Given a list of pci index's init the pci chip and call reset on it"""

    reset_wh_pci_idx = []
    reset_gs_devs = []
    reset_bh_pci_idx = []
    for pci_idx in list_of_boards:
        try:
            chip = PciChip(pci_interface=pci_idx)
        except Exception as e:
            print(
                CMD_LINE_COLOR.RED,
                f"Error accessing board at pci index {pci_idx}! Use -ls to see all devices available to reset",
                CMD_LINE_COLOR.ENDC,
            )
        if chip.as_wh():
            reset_wh_pci_idx.append(pci_idx)
        elif chip.as_gs():
            reset_gs_devs.append(chip)
        elif chip.as_bh():
            reset_bh_pci_idx.append(pci_idx)
        else:
            print(f"{CMD_LINE_COLOR.RED}Unknown chip!!{CMD_LINE_COLOR.ENDC}")
            sys.exit(1)

    # reset wh devices with pci indices
    if len(reset_wh_pci_idx) > 0:
        WHChipReset().full_lds_reset(pci_interfaces=reset_wh_pci_idx, silent=True)

    if len(reset_bh_pci_idx) > 0:
        BHChipReset().full_lds_reset(pci_interfaces=reset_bh_pci_idx, silent=True)

    # reset gs devices by creating a partially init backend
    if len(reset_gs_devs) > 0:
        for i, device in enumerate(reset_gs_devs):
            GSTensixReset(device).tensix_reset(silent=True)

    if reinit:
        # Enable backtrace for debugging
        os.environ["RUST_BACKTRACE"] = "full"

        print(
            CMD_LINE_COLOR.PURPLE,
            f"Re-initializing boards after reset....",
            CMD_LINE_COLOR.ENDC,
        )
        try:
            chips = detect_chips_with_callback()
        except Exception as e:
            print(
                CMD_LINE_COLOR.RED,
                f"Error when re-initializing chips!\n {e}",
                CMD_LINE_COLOR.ENDC,
            )
            sys.exit(1)


def pci_indices_from_json(json_dict):
    """Parse pci_list from reset json"""
    pci_indices = []
    reinit = False
    if "gs_tensix_reset" in json_dict.keys():
        pci_indices.extend(json_dict["gs_tensix_reset"]["pci_index"])
    if "wh_link_reset" in json_dict.keys():
        pci_indices.extend(json_dict["wh_link_reset"]["pci_index"])
    if "re_init_devices" in json_dict.keys():
        reinit = json_dict["re_init_devices"]
    return pci_indices, reinit


def mobo_reset_from_json(json_dict) -> dict:
    """Parse pci_list from reset json and init mobo reset"""
    if "wh_mobo_reset" in json_dict.keys():
        mobo_dict_list = []
        for mobo_dict in json_dict["wh_mobo_reset"]:
            # Only add the mobos that have a name
            if "MOBO NAME" not in mobo_dict["mobo"]:
                mobo_dict_list.append(mobo_dict)
        # If any mobos - do the reset
        if mobo_dict_list:
            GalaxyReset().warm_reset_mobo(mobo_dict_list)
            # If there are mobos to reset, remove link reset pci index's from the json
            try:
                wh_link_pci_indices = json_dict["wh_link_reset"]["pci_index"]
                for entry in mobo_dict_list:
                    if "nb_host_pci_idx" in entry.keys() and entry["nb_host_pci_idx"]:
                        # remove the list of WH pcie index's from the reset list
                        wh_link_pci_indices = list(
                            set(wh_link_pci_indices) - set(entry["nb_host_pci_idx"])
                        )
                json_dict["wh_link_reset"]["pci_index"] = wh_link_pci_indices
            except Exception as e:
                print(
                    CMD_LINE_COLOR.RED,
                    f"Error! {e}",
                    CMD_LINE_COLOR.ENDC,
                )

    return json_dict


def parse_reset_input(value):
    """Validate the reset inputs - either list of int pci IDs or a json config file"""
    if not value:
        return None
    try:
        # Attempt to parse as a JSON file
        with open(value, "r") as json_file:
            data = json.load(json_file)
            return data
    except json.JSONDecodeError as e:
        print(
            CMD_LINE_COLOR.RED,
            f"Please check the format of the json file.\n {e}",
            CMD_LINE_COLOR.ENDC,
        )
        sys.exit(1)
    except FileNotFoundError:
        # If no file found, attempt to parse as a list of comma separated integers
        print(
            CMD_LINE_COLOR.YELLOW,
            "File not found!\n To generate a reset json config file run tt-smi -g",
            CMD_LINE_COLOR.ENDC,
        )
        return None


def print_all_available_devices(devices):
    """Print all available boards on host"""
    console = get_console()
    table = Table()
    table.add_column("Pci Dev ID")
    table.add_column("Board Type")
    table.add_column("Device Series")
    table.add_column("Board Number")
    table.add_column("Coordinates")
    for i, device in enumerate(devices):
        chip = device.luwen_chip
        board_id = hex(device.board_id()).replace("0x", "")
        board_type = get_board_type(board_id)
        device_series = device.arch()
        pci_dev_id = device.interface_id if not device.is_remote else "N/A"
        coords = device.coord()
        if isinstance(chip, WhChip):
            suffix = " R" if device.is_remote else " L"
            board_type = board_type + suffix

        table.add_row(
            f"{pci_dev_id}",
            f"{device_series}",
            f"{board_type}",
            f"{board_id}",
            f"{coords}",
        )
    console.print(table)

def get_board_type(board_id: str) -> str:
    """
    Get board type from board ID string.
    Ex:
        Board ID: AA-BBBBB-C-D-EE-FF-XXX
                   ^     ^ ^ ^  ^  ^   ^
                   |     | | |  |  |   +- XXX
                   |     | | |  |  +----- FF
                   |     | | |  +-------- EE
                   |     | | +----------- D
                   |     | +------------- C = Revision
                   |     +--------------- BBBBB = Unique Part Identifier (UPI)
                   +--------------------- AA
    """
    serial_num = int(f"0x{board_id}", base=16)
    upi = (serial_num >> 36) & 0xFFFFF

    # Grayskull cards
    if upi == 0x3:
        return "e150"
    elif upi == 0xA:
        return "e300"
    elif upi == 0x7:
        return "e75"

    # Wormhole cards
    elif upi == 0x8:
        return "nb_cb"
    elif upi == 0xB:
        return "wh_4u"
    elif upi == 0x14:
        return "n300"
    elif upi == 0x18:
        return "n150"
    elif upi == 0x35:
        return "tt-galaxy-wh"

    # Blackhole cards
    elif upi == 0x36:
        return "bh-scrappy"
    elif upi == 0x43:
        return "p100a"
    elif upi == 0x40:
        return "p150a"
    elif upi == 0x41:
        return "p150b"
    else:
        return "N/A"

def prefix_color_picker(current_value, max_value):
    if current_value < max_value * 0.85:
        return "[green]"
    else:
        return "[orange3]"

def asic_temperature_parser(temp, dev):
    """ASIC temperature is reported with different schema for BH vs other chips"""
    if dev.as_bh():
        # BH temp is reported as signed 16_16 integer that needs to be split into two 16 bit values
        return (temp >> 16) + (temp & 0xFFFF) / 65536.0
    else:
        return (temp & 0xFFFF) / 16

def timed_wait(seconds):
    """Wait for a specified number of seconds, printing the progress."""
    print("\033[93mWaiting for {} seconds: 0\033[0m".format(seconds), end='')
    sys.stdout.flush()

    for i in range(1, seconds + 1):
        time.sleep(1)
        # Move cursor back and overwrite the number
        print("\r\033[93mWaiting for {} seconds: {}\033[0m".format(seconds, i), end='')
        sys.stdout.flush()
    print()

def reset_6u_glx():
    """Reset WH Galaxy trays and detect chips post reset."""
    print(
        CMD_LINE_COLOR.PURPLE,
        f"Resetting WH Galaxy trays with reset command...",
        CMD_LINE_COLOR.ENDC,
    )
    run_wh_ubb_ipmi_reset(ubb_num="0xF", dev_num="0xFF", op_mode="0x0", reset_time="0xF")
    timed_wait(30)
    run_ubb_wait_for_driver_load()
    print(
        CMD_LINE_COLOR.PURPLE,
        f"Re-initializing boards after reset....",
        CMD_LINE_COLOR.ENDC,
    )
    try:
        devs = detect_chips_fallible(
            local_only=True,
            continue_on_failure=False,
            callback=None,
            noc_safe=True,
        )
        print(
            CMD_LINE_COLOR.GREEN,
            f"Re-initialized {len(devs)} chips after reset.",
            CMD_LINE_COLOR.ENDC,
        )
    except Exception as e:
        print(
            CMD_LINE_COLOR.RED,
            f"Error when re-initializing chips!\n {e}",
            CMD_LINE_COLOR.ENDC,
        )
        # Error out if chips don't initalize
    return

def generate_table(devices) -> Table:
    """Make a table to display telemetry values."""
    table = Table(
        title=" ",
    )
    table.add_column("ID")
    table.add_column("Core Voltage (V)")
    table.add_column("Core Current (A)")
    table.add_column("AICLK (MHz)")
    table.add_column("Power (W)")
    table.add_column("Core Temp (°C)")

    for i, dev in enumerate(devices):
        telem = jsons.dump(dev.get_telemetry())
        current = int(hex(telem["tdc"]), 16) & 0xFFFF
        voltage = int(hex(telem["vcore"]), 16) / 1000
        aiclk = int(hex(telem["aiclk"]), 16) & 0xFFFF
        power = int(hex(telem["tdp"]), 16) & 0xFFFF
        asic_temperature = asic_temperature_parser(int(hex(telem["asic_temperature"]), 16), dev)
        vdd_max = int(hex(telem["vdd_limits"]), 16) >> 16
        curr_limit = int(hex(telem["tdc"]), 16) >> 16
        aiclk_limit = int(hex(telem["aiclk"]), 16) >> 16
        pwr_limit = int(hex(telem["tdp"]), 16) >> 16
        thm_limit = int(hex(telem["thm_limits"]), 16) & 0xFFFF
        table.add_row(
            f"{i}",
            f"{voltage:4.2f}[light_goldenrod1] / {vdd_max/1000:4.2f}",
            f"{prefix_color_picker(current, curr_limit)}{current:5.1f}[light_goldenrod1] / {curr_limit:5.1f}",
            f"{prefix_color_picker(aiclk, aiclk_limit)}{aiclk:4.0f}[light_goldenrod1] / {aiclk_limit:4.0f}",
            f"{prefix_color_picker(power, pwr_limit)}{power:5.1f}[light_goldenrod1] / {pwr_limit:5.1f}",
            f"{prefix_color_picker(asic_temperature, thm_limit)}{asic_temperature:4.1f}[light_goldenrod1] / {thm_limit:4.1f}",
        )

    return table



#!/usr/bin/env python3
"""
Compatibility setup.py for Ubuntu 22.04 packaging tools.
Uses setuptools_scm for dynamic versioning from git tags.
"""
from setuptools import setup, find_packages

if __name__ == "__main__":
    setup(
        # Fallback for older setuptools versions
        name="tt-burnin",
        use_scm_version=True,
        packages=find_packages(),
        python_requires=">=3.10",
        setup_requires=['setuptools_scm'],
    )



# SPDX-FileCopyrightText: © 2024 Tenstorrent AI ULC
# SPDX-License-Identifier: Apache-2.0
"""
Tenstorrent Burnin (TT-Burnin) is a command line utility
to run a high power consumption workload on TT devices.
"""
from __future__ import annotations

import os
import sys
import time
import argparse
import threading
import tt_burnin
from rich.live import Live
from rich.text import Text
from rich.console import Group
from importlib.resources import path
from tt_burnin.chip import GsChip, WhChip, RemoteWhChip, BhChip
from tt_burnin.load_ttx import load_ttx_file, TtxFile, CoreId
from tt_tools_common.ui_common.themes import CMD_LINE_COLOR
from tt_burnin.utils import (
    parse_reset_input,
    mobo_reset_from_json,
    pci_indices_from_json,
    pci_board_reset,
    print_all_available_devices,
    generate_table,
    get_board_type,
    reset_6u_glx,

)
from tt_tools_common.utils_common.tools_utils import (
    detect_chips_with_callback,
)


def reset_all_devices(devices, reset_filename=None):
    """Reset all devices"""
    print(CMD_LINE_COLOR.BLUE, "Resetting devices on host...", CMD_LINE_COLOR.ENDC)
    LOG_FOLDER = os.path.expanduser("~/.config/tenstorrent")
    log_filename = f"{LOG_FOLDER}/reset_config.json"
    if not devices:
        print(
            CMD_LINE_COLOR.RED,
            "No devices detected. Exiting...",
            CMD_LINE_COLOR.ENDC,
        )
        sys.exit(1)
    # Check board type and reset accordingly
    board_id = hex(devices[0].board_id()).replace("0x", "")
    board_type = get_board_type(board_id)
    if board_type == "tt-galaxy-wh":
        # Perform a full galaxy reset and detect chips post reset
        reset_6u_glx()
        return

    # If input is just reset board
    if not reset_filename:
        log_filename = reset_filename
    data = parse_reset_input(log_filename)
    if data:
        # reset using the json file
        parsed_dict = mobo_reset_from_json(data)
        pci_indices, reinit = pci_indices_from_json(parsed_dict)
        if pci_indices:
            pci_board_reset(pci_indices, reinit)
    else:
        # reset all boards
        dev_ids = []
        for device in devices:
            if not device.is_remote():
                dev_ids.append(device.get_pci_interface_id())
        pci_board_reset(dev_ids, reinit=True)


def start_burnin_gs(
    device,
    keep_trisc_under_reset: bool = False,
    stagger_start: bool = False,
    no_check: bool = False,
    idle: bool = False
):
    BRISC_SOFT_RESET = 1 << 11
    TRISC_SOFT_RESETS = (1 << 12) | (1 << 13) | (1 << 14)
    NCRISC_SOFT_RESET = 1 << 18
    STAGGERED_START_ENABLE = (1 << 31) if stagger_start else 0

    device.noc_broadcast32(
        0, 0xFFB121B0, BRISC_SOFT_RESET | TRISC_SOFT_RESETS | NCRISC_SOFT_RESET
    )

    # Deassert riscv reset
    device.arc_msg(0xBA)

    # Go busy
    device.arc_msg(0x52)

    if not idle:
        with path("tt_burnin", "") as data_path:
            load_ttx_file(
                device,
                TtxFile(str(data_path.joinpath("ttx/gspv.ttx"))),
                {CoreId(0, 0): device.get_tensix_locations()},
                no_check,
            )

    if keep_trisc_under_reset:
        soft_reset_value = (
            NCRISC_SOFT_RESET | TRISC_SOFT_RESETS | STAGGERED_START_ENABLE
        )
    else:
        soft_reset_value = NCRISC_SOFT_RESET | STAGGERED_START_ENABLE

    # Take cores out of reset
    device.noc_broadcast32(0, 0xFFB121B0, soft_reset_value)


def stop_burnin_gs(device):
    BRISC_SOFT_RESET = 1 << 11
    TRISC_SOFT_RESETS = (1 << 12) | (1 << 13) | (1 << 14)
    NCRISC_SOFT_RESET = 1 << 18

    # Go idle
    device.arc_msg(0x54)

    # Put tensix back under soft reset
    device.noc_broadcast32(
        0, 0xFFB121B0, BRISC_SOFT_RESET | TRISC_SOFT_RESETS | NCRISC_SOFT_RESET
    )


def start_burnin_wh(
    device,
    keep_trisc_under_reset: bool = False,
    stagger_start: bool = False,
    no_check: bool = False,
    idle: bool = False
):
    BRISC_SOFT_RESET = 1 << 11
    TRISC_SOFT_RESETS = (1 << 12) | (1 << 13) | (1 << 14)
    NCRISC_SOFT_RESET = 1 << 18
    STAGGERED_START_ENABLE = (1 << 31) if stagger_start else 0

    # Put tensix under soft reset
    device.noc_broadcast32(
        0, 0xFFB121B0, BRISC_SOFT_RESET | TRISC_SOFT_RESETS | NCRISC_SOFT_RESET
    )

    # Deassert riscv reset
    device.arc_msg(0xBA)

    # Go busy
    device.arc_msg(0x52)

    if not idle:
        with path("tt_burnin", "") as data_path:
            load_ttx_file(
                device,
                TtxFile(str(data_path.joinpath("ttx/whpv.ttx"))),
                {CoreId(0, 0): device.get_tensix_locations()},
                no_check,
            )

    if keep_trisc_under_reset:
        soft_reset_value = (
            NCRISC_SOFT_RESET | TRISC_SOFT_RESETS | STAGGERED_START_ENABLE
        )
    else:
        soft_reset_value = NCRISC_SOFT_RESET | STAGGERED_START_ENABLE

    # Take cores out of reset
    device.noc_broadcast32(0, 0xFFB121B0, soft_reset_value)


def stop_burnin_wh(device):
    BRISC_SOFT_RESET = 1 << 11
    TRISC_SOFT_RESETS = (1 << 12) | (1 << 13) | (1 << 14)
    NCRISC_SOFT_RESET = 1 << 18

    # Go idle
    device.arc_msg(0x54)

    # Put tensix back under soft reset
    device.noc_broadcast32(
        0, 0xFFB121B0, BRISC_SOFT_RESET | TRISC_SOFT_RESETS | NCRISC_SOFT_RESET
    )


def start_burnin_bh(
    device,
    keep_trisc_under_reset: bool = False,
    stagger_start: bool = False,
    no_check: bool = False,
    idle: bool = False
):
    BRISC_SOFT_RESET = 1 << 11
    TRISC_SOFT_RESETS = (1 << 12) | (1 << 13) | (1 << 14)
    NCRISC_SOFT_RESET = 1 << 18
    STAGGERED_START_ENABLE = (1 << 31) if stagger_start else 0

    # Put tensix under soft reset
    device.noc_broadcast32(
        0, 0xFFB121B0, BRISC_SOFT_RESET | TRISC_SOFT_RESETS | NCRISC_SOFT_RESET
    )

    # Go busy
    device.arc_msg(0x52)

    if not idle:
        with path("tt_burnin", "") as data_path:
            load_ttx_file(
                device,
                TtxFile(str(data_path.joinpath("ttx/bhpv.ttx"))),
                {CoreId(0, 0): device.get_tensix_locations()},
                no_check
            )

    if keep_trisc_under_reset:
        soft_reset_value = (
            NCRISC_SOFT_RESET | TRISC_SOFT_RESETS | STAGGERED_START_ENABLE
        )
    else:
        soft_reset_value = NCRISC_SOFT_RESET | STAGGERED_START_ENABLE

    # Take cores out of reset
    device.noc_broadcast32(0, 0xFFB121B0, soft_reset_value)


def stop_burnin_bh(device):
    BRISC_SOFT_RESET = 1 << 11
    TRISC_SOFT_RESETS = (1 << 12) | (1 << 13) | (1 << 14)
    NCRISC_SOFT_RESET = 1 << 18

    # Go idle
    device.arc_msg(0x54)

    # Put tensix back under soft reset
    device.noc_broadcast32(
        0, 0xFFB121B0, BRISC_SOFT_RESET | TRISC_SOFT_RESETS | NCRISC_SOFT_RESET
    )


def parse_args():
    # Parse arguments
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=tt_burnin.__version__,
    )
    parser.add_argument(
        "--reset_file",
        type=parse_reset_input,
        metavar="reset_config.json",
        default=None,
        help=(
            "Provide a custom reset json file for the host."
            "Generate a default reset json file with the -g option with tt-smi."
        ),
        dest="reset",
    )
    parser.add_argument(
        "--no-reset",
        action="store_true",
        default=False,
        help="Don't issue a reset before or after burning (WARNING: This may cause burnin or your next workload to no longer function)",
    )
    parser.add_argument(
        "--no-check",
        action="store_true",
        default=False,
        help="Don't check tensix fw after loading (WARNING: if the workload was loaded incorrectly burnin may not run at maximum load)",
    )
    parser.add_argument(
        "--idle",
        action="store_true",
        default=False,
        help="Don't load the power virus workload, just run the tensix idle",
    )
    # subparsers = parser.add_subparsers(title="command", dest="command", required=True)
    return parser.parse_args()


def main():
    args = parse_args()
    os.environ["RUST_BACKTRACE"] = "full"
    # Allow non blocking read for accepting user input before stopping burnin
    os.set_blocking(sys.stdin.fileno(), False)
    all_devices = detect_chips_with_callback()
    devs = []
    devices = []
    for device in all_devices:
        if device.as_gs() is not None:
            devs.append(GsChip(device.as_gs()))
        elif device.as_wh() is not None:
            if device.is_remote():
                devs.append(RemoteWhChip(device.as_wh()))
            else:
                devs.append(WhChip(device.as_wh()))
        elif device.as_bh() is not None:
            devs.append(BhChip(device.as_bh()))
        else:
            raise ValueError("Did not recognize board")
        devices.append(device)
    print_all_available_devices(devs)
    if not args.no_reset:
        reset_all_devices(devices, reset_filename=args.reset)

    try:
        print()
        print(
            CMD_LINE_COLOR.BLUE,
            "Starting TT-Burnin workload on all boards. WARNING: Opening SMI might cause unexpected behavior",
            CMD_LINE_COLOR.ENDC,
        )
        print()
        def start_burnin(device, idx, total):
                print(
                    CMD_LINE_COLOR.PURPLE,
                    f"Starting TT-Burnin workload on device {idx + 1}/{total}",
                    CMD_LINE_COLOR.ENDC,
                )
                if isinstance(device, GsChip):
                    start_burnin_gs(device)
                elif isinstance(device, WhChip):
                    start_burnin_wh(device)
                elif isinstance(device, BhChip):
                    start_burnin_bh(device)
                else:
                    raise NotImplementedError(f"Don't support {device}")

        # Thread the start of burnin for faster speed
        threads = []
        for i, device in enumerate(devs):
            t = threading.Thread(target=start_burnin, args=(device, i, len(devs)))
            t.start()
            threads.append(t)
        for t in threads:
            t.join()

        text = Text(
            " Press Enter to STOP TT-Burnin on all boards...", style="bold yellow"
        )

        # Create a live update for telemetry widget
        with Live(Group(generate_table(devices), text), refresh_per_second=10) as live:
            while True:
                # Break if there is any user keypress
                c = sys.stdin.read(1)
                if len(c) > 0:
                    break
                live.update(Group(generate_table(devices), text))
                time.sleep(0.1)
    except Exception as e:
        import traceback
        traceback.print_exc()
        print(e)
    finally:
        print()
        print(
            CMD_LINE_COLOR.GREEN,
            "Stopping TT-Burnin workload on all boards.",
            CMD_LINE_COLOR.ENDC,
        )
        print()
        # Thread the stop of burnin for faster speed
        def stop_burnin(device):
            if isinstance(device, GsChip):
                stop_burnin_gs(device)
            elif isinstance(device, WhChip):
                stop_burnin_wh(device)
            elif isinstance(device, BhChip):
                stop_burnin_bh(device)
            else:
                raise NotImplementedError(f"Don't support {device}")

        stop_threads = []
        for device in devs:
            t = threading.Thread(target=stop_burnin, args=(device,))
            t.start()
            stop_threads.append(t)
        for t in stop_threads:
            t.join()

        # Final reset to restore state
        if not args.no_reset:
            reset_all_devices(devices, reset_filename=args.reset)
